name = "director"

from os import chdir
import logging
import json
import traceback

logger = logging.getLogger(__package__)
import importlib.util

from behave.__main__ import run_behave
from behave.__main__ import Configuration

from .formats import register_formats
from .mocking import MixinBase, LogMixin, RelayLog, MockMixin
from .mocking import VacantLog, RelayLog, MockLog
from .identify import WidgetSelector, find_widgets
from director.formatters import GradescopeFormatter


def setup(context):
    context.under_test = context.config.under_test
    register_formats()


def load_under_test(path):
    spec = importlib.util.spec_from_file_location("under_test", path)
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)
    return foo


def test(tests, target, working_directory=".", gradescope=False):
    extra_args = [] if not gradescope else ["--no-summary"]
    config = Configuration(command_args=[tests, "--no-source", "--no-timings"] + extra_args)
    config.steps_dir = "."

    try:
        config.under_test = load_under_test(target)
    except SyntaxError as e:
        if not gradescope:
            raise e
        
        print(json.dumps({
            "score": 0,
            "output": traceback.format_exc()
        }))
        return

    if gradescope:
        config.default_format = "gradescope"
        config.more_formatters = {"gradescope": GradescopeFormatter}
        config.setup_formats()

    chdir(working_directory)

    run_behave(config)        


__export__ = [MixinBase, LogMixin, RelayLog, MockMixin, VacantLog, RelayLog, MockLog, WidgetSelector, setup]
__test__ = {obj.__name__ : obj for obj in __export__}
__all__ = [cls.__name__ for cls in __export__] + ["logger", "find_widgets"]



import json
from behave.formatter.base import Formatter
from behave.model import Scenario, Step
from behave.model_core import Status


class GradescopeFormatter(Formatter):
    # def feature(self, feature):
    #     print(feature)

    def __init__(self, stream_opener, config):
        super().__init__(stream_opener, config)

        self._tests = []
        self._results = {
            "tests": self._tests
        }

        self.reset(None)

    def reset(self, scenario):
        self._current_scenario: Scenario = scenario
        self._passed = True
        self._output = ""

    def _make_test(self):
        return {
            "score": 1 if self._passed else 0,
            "max_score": 1,
            "name": f"Scenario: {self._current_scenario.name}",
            "output": self._output,
            "visibility": "visible",
        }
    
    def scenario(self, scenario):
        if self._current_scenario is not None:
            self._tests.append(self._make_test())
        self.reset(scenario)

    def result(self, step: Step):
        if step.status == Status.passed:
            self._output += f"{step.keyword} {step.name}\n"
            return

        self._passed = False

        if step.status == Status.skipped:
            self._output += f"{step.keyword} {step.name} (skipped)\n"
            return

        self._output += f"""{step.keyword} {step.name}
    status: {step.status}
    {step.error_message}

    output: {step.captured.output}
"""

        

    def eof(self):
        print()

    def close(self):
        self.stream.write(json.dumps(self._results))
        super().close()

    # def step(self, step):
    #     print(step)



"""
Identify GUI widgets based on the functional characteristics
within an existing model.

Allows testing to occur in an environment where no knowledge
is known about the implementation details of the GUI.
"""
import tkinter as tk
from typing import List
from . import logger


def _widget_selector(parent, selector):
    if selector(parent):
        yield parent
    for child in parent.children.values():
    #for child in parent.winfo_children(): doesn't work as students override self._root
        yield from _widget_selector(child, selector)


def find_widgets(selector, widget) -> List[tk.Widget]:
    return list(_widget_selector(widget, selector))


class WidgetSelector:
    @staticmethod
    def aggregate(*selectors):
        def cb(widget):
            for selector in selectors:
                if selector(widget):
                    return True
            return False

        return cb

    @staticmethod
    def by_text(expected):
        def f(widget):
            try:
                actual = widget.cget("text")
                return actual == expected
            except tk.TclError:
                return False
        
        return f

    @staticmethod
    def by_rough_text(expected):
        def f(widget):
            try:
                actual = widget.cget("text").lower().strip()
                return actual == expected
            except tk.TclError:
                return False
        
        return f

    @staticmethod
    def by_label(expected):
        def f(widget):
            try:
                actual = widget.cget("label").lower()
                print(actual)
                return expected in actual
            except tk.TclError:
                return False

        return f

    @staticmethod
    def has_text():
        def f(widget):
            try:
                widget.cget("text")
                return True
            except tk.TclError:
                return False
        return f

    @staticmethod
    def by_type(expected):
        def f(widget):
            return isinstance(widget, expected)
        return f

    @staticmethod
    def is_leaf():
        def f(widget):
            return not widget.children
        return f

    @staticmethod
    def all():
        def f(widget):
            return True
        return f



from . import logger
from behave import register_type
import parse
from enum import Enum


class RelativePosition(Enum):
    Left = 0
    Right = 1
    Above = 2
    Below = 3


@parse.with_pattern(r"(\w|\s|\d|.)*")
def parse_string(text):
    logger.debug(f"Parsing Text pattern: {text}")
    return text


@parse.with_pattern(r"\d+")
def parse_number(text):
    logger.debug(f"Parsing Number pattern: {text}")
    return int(text)


@parse.with_pattern(r"left of|right of|above|below")
def parse_relative_position(text):
    logger.debug(f"Parsing RelativePosition pattern: {text}")
    if text == "left of":
        return RelativePosition.Left
    elif text == "right of":
        return RelativePosition.Right
    elif text == "above":
        return RelativePosition.Above
    elif text == "below":
        return RelativePosition.Below
    return None # not reached


def register_formats():
    logger.debug("registering Text pattern")
    register_type(Text=parse_string)
    logger.debug("registering Number pattern")
    register_type(Number=parse_number)
    logger.debug("registering RelativePosition pattern")
    register_type(RelativePosition=parse_relative_position)



"""
Mock core components of the tkinter library.

Allows for automated GUI events, such as key presses or
time steps, to be triggered.
"""
import types
import functools
import inspect


def copy_function(f):
    """
    Perform a deep copy of a python function.
    
    Based on http://stackoverflow.com/a/6528148/190597 (Glenn Maynard)
    """
    g = types.FunctionType(f.__code__, f.__globals__, name=f.__name__,
                           argdefs=f.__defaults__,
                           closure=f.__closure__)
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    return g


class MixinBase(object):
    """
    Core mocking functionality, allows calls to methods to be replaced with
    no-op behaviour.

    Examples:
    >>> class MockMe:
    ...     def do_it(self):
    ...         print("called")
    >>> mockme = MockMe()
    >>> mockme.do_it()
    called
    >>> mock = MixinBase(MockMe, "do_it")
    >>> mockme.do_it()
    >>> mock.restore()
    >>> mockme.do_it()
    called
    """
    def __init__(self, context, reference) -> None:
        assert hasattr(context, reference)
        self._commands = []
        callable = getattr(context, reference)
        self._original = copy_function(callable)
        self._context, self._reference = context, reference
        setattr(self._context, self._reference, self._call)

        for mixin in self.__class__.__mro__:
            if mixin == self.__class__:
                continue
            if hasattr(mixin, "setup"):
                mixin.setup(self)

    def _call(self, *args, **kwargs):
        for mixin in self.__class__.__mro__:
            if mixin == self.__class__:
                continue
            if hasattr(mixin, "inject"):
                mixin.inject(self, *args, **kwargs)
    
    def restore(self):
        setattr(self._context, self._reference, self._original)


class LogMixin(MixinBase):
    """
    Log calls to mocked references.

    >>> class Logger(LogMixin):
    ...     pass
    >>> class MockMe:
    ...     def do_it(self, *args, **kwargs):
    ...         print("called")
    >>> mockme = MockMe()
    >>> mockme.do_it()
    called
    >>> mock = Logger(MockMe, "do_it")
    >>> mockme.do_it("life", answer=42)
    >>> mockme.do_it()
    >>> mock.restore()
    >>> mockme.do_it()
    called
    >>> mock.logs
    [(('life',), {'answer': 42}), ((), {})]
    """
    def setup(self) -> None:
        self._records = []

    @property
    def logs(self):
        return self._records[:]

    def inject(self, *args, **kwargs):
        self._records.append((args, kwargs))


class MockMixin(MixinBase):
    """
    Inject arbitrary calls to a mocked reference.

    >>> class Injection(MockMixin):
    ...     pass
    >>> class MockMe:
    ...     def do_it(self, *args, **kwargs):
    ...         print("called")
    >>> mockme = MockMe()
    >>> mockme.do_it()
    called
    >>> mock = Injection(MockMe, "do_it")
    >>> mock.register(lambda *a, **kw: print("mocked", a, kw))
    >>> mockme.do_it("life", answer=42)
    mocked ('life',) {'answer': 42}
    >>> mockme.do_it()
    mocked () {}
    >>> mock.restore()
    >>> mockme.do_it()
    called
    """
    def setup(self) -> None:
        self._mocks = []

    def register(self, mock):
        self._mocks.append(mock)
        
    def inject(self, *args, **kwargs):
        for mock in self._mocks:
            mock(*args, **kwargs)


class RelayMixin(MixinBase):
    """
    Relay mocked calls to the original reference.

    >>> class Relay(RelayMixin):
    ...     pass
    >>> class MockMe:
    ...     def do_it(self, *args, **kwargs):
    ...         print("called")
    >>> mockme = MockMe()
    >>> mockme.do_it()
    called
    >>> mock = Relay(MockMe, "do_it")
    >>> mockme.do_it("life", answer=42)
    called
    >>> mockme.do_it()
    called
    >>> mock.restore()
    >>> mockme.do_it()
    called
    >>> mock.logs
    [(('life',), {'answer': 42}), ((), {})]
    """
    def inject(self, *args, **kwargs):
        self._original(self, *args, **kwargs)


# a couple of useful example mocking classes

class VacantLog(LogMixin):
    pass


class RelayLog(LogMixin, RelayMixin):
    pass


class MockLog(LogMixin, MockMixin):
    pass



import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="director-tester",
    version="0.0.1",
    packages=[
        'director',
    ],
    install_requires=["behave", "Pillow"],
    description="Perform automated testing of Tkinter GUI applications",
    long_description=README,
    long_description_content_type="text/markdown",
    # url="https://github.com/BraeWebb/director",
    author="Brae Webb",
    author_email="b.webb@uq.edu.au",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
)


from behave import *
from .identify import *
from .formats import RelativePosition, register_formats

register_formats()


@then("I see text displaying, roughly, \"{text:Text}\"")
def rough_text(context, text):
    search_for = text.lower().strip()
    widgets = find_widgets(WidgetSelector.by_rough_text(search_for), context.window)
    assert len(widgets) == 1, f"cannot find exactly one widget roughly matching the text \"{text}\", found {widgets}"
    context.last = widgets[0]


@then("I see text displaying, exactly, {text:Text}")
def exact_text(context, text):
    widgets = find_widgets(WidgetSelector.by_text(text), context.window)
    assert len(widgets) == 1, f"cannot find exactly one widget exactly matching the text \"{text}\", found {widgets}"
    context.last = widgets[0]


@then("it is {position:RelativePosition} all other widgets")
def relative_to_all(context, position):
    widgets = find_widgets(WidgetSelector.all(), context.window)
    it: tk.Widget = context.last
    its_position = it.winfo_x(), it.winfo_y()

    for widget in widgets:
        position = widget.winfo_x(), it.winfo_y()
        if position == RelativePosition.Left:
            assert its_position[0] < position[0], f"{widget} is further left than {it}"
        elif position == RelativePosition.Right:
            assert its_position[0] > position[0], f"{widget} is further right than {it}"
        elif position == RelativePosition.Above:
            assert its_position[1] < position[1], f"{widget} is above {it}"
        elif position == RelativePosition.Below:
            assert its_position[1] > position[1], f"{widget} is below {it}"
